/* $Header: /cvsroot/dhcp-agent/dhcp-agent/src/dhcp-server-conf.c,v 1.8 2003/07/15 10:56:34 actmodern Exp $
 * 
 * Copyright 2002 Thamer Alharbash <tmh@whitefang.com>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 3. The names of the authors may not be used to endorse or promote
 * products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#define MODULE_NAME "dhcp-server-conf"

#include "dhcp-local.h"
#include "dhcp-limits.h"
#include "dhcp-libutil.h"
#include "dhcp-librawnet.h"
#include "dhcp-option.h"
#include "dhcp-lease.h"

#include "dhcp-conf.h"
#include "dhcp-conf-var.h"
#include "dhcp-server-conf.h"
#include "dhcp-server-defaults.h"
#include "dhcp-options-strings.h"

static const char *var_strings[] = {
    "default-renew-percent",
    "default-rebind-percent",
    "poll-timeout"
};

static const arg_symbol_t var_symbols[] = {
    SERVER_VAR_RENEW_PERCENT,
    SERVER_VAR_REBIND_PERCENT,
    SERVER_VAR_POLL_TIMEOUT,
};

static const arg_type_t     var_arg_types[] = { CONF_IDENTIFIER, CONF_ASSIGNMENT, CONF_STRING };
static const char         **var_arg_strings[] = { var_strings, NULL, NULL };
static const arg_symbol_t  *var_arg_symbols[] = { var_symbols, NULL, NULL };

static const arg_type_t     range_lease_arg_types[] = {  CONF_ADDRESS, CONF_ADDRESS, CONF_GROUP};
static const char         **range_lease_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *range_lease_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     hostname_lease_arg_types[] = {  CONF_STRING, CONF_ADDRESS, CONF_GROUP};
static const char         **hostname_lease_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *hostname_lease_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     mac_lease_arg_types[] = {  CONF_MAC_ADDRESS, CONF_ADDRESS, CONF_GROUP};
static const char         **mac_lease_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *mac_lease_arg_symbols[] = { NULL, NULL, NULL };

static const arg_type_t     option_arg_types[] = {  CONF_STRING, CONF_ASSIGNMENT, CONF_STRING_LIST };
static const char         **option_arg_strings[] = { NULL, NULL, NULL };
static const arg_symbol_t  *option_arg_symbols[] = { NULL, NULL, NULL };

/* command structures. */

/* set command. */
static const command_t command_set = {
    DIRECTIVE_SET,
    "set",
    3,
    var_arg_strings,
    var_arg_types,
    var_arg_symbols,
};

/* range-lease command. */
static const command_t command_range_lease = {
    DIRECTIVE_RANGE_LEASE,
    "range-lease",
    3,
    range_lease_arg_strings,
    range_lease_arg_types,
    range_lease_arg_symbols,
};

/* hostname-lease command. */
static const command_t command_hostname_lease = {
    DIRECTIVE_HOSTNAME_LEASE,
    "hostname-lease",
    3,
    hostname_lease_arg_strings,
    hostname_lease_arg_types,
    hostname_lease_arg_symbols,
};

/* mac-lease command. */
static const command_t command_mac_lease = {
    DIRECTIVE_MAC_LEASE,
    "mac-lease",
    3,
    mac_lease_arg_strings,
    mac_lease_arg_types,
    mac_lease_arg_symbols,
};

static const command_t command_option = {
    DIRECTIVE_OPTION,
    "option",
    3,
    option_arg_strings,
    option_arg_types,
    option_arg_symbols,
};

static const command_t *commands[] = {
    &command_set,
    &command_range_lease,
    &command_hostname_lease,
    &command_mac_lease,
    &command_option,
    NULL,
};

/* forward declaration of directive handlers. */

static int directive_set_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data);
static int directive_range_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data);
static int directive_option_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data);
static int directive_hostname_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data);
static int directive_hw_address_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data);

/* indexed by directive types in dhcp-client-conf.h . */
directive_handler_t directive_handlers[] = {
    directive_set_handler,
    directive_range_handler,
    directive_hostname_handler,
    directive_hw_address_handler,
    directive_option_handler,
};

/* accessors. */

uint16_t server_conf_get_poll_timeout(server_conf_t *sc)
{
    return sc->poll_timeout;
}


/* * * * * * * * * * *
 * utility routines. *
 * * * * * * * * * * */

/* get name of configuration file for server. */
static char *get_conf_options_fname(const char *interface)
{
    char *fname;
    stringbuffer_t *sb;

    sb = stringbuffer_create();
    stringbuffer_aprintf(sb, "%s/default.conf", DHCPSYSCONF_SERVERDIR);

    if(file_exists(stringbuffer_getstring(sb))) {

        fname = xstrdup(stringbuffer_getstring(sb));
        stringbuffer_destroy(sb);
        return fname;

    }

    stringbuffer_clear(sb);
    stringbuffer_aprintf(sb, "%s/%s.conf", DHCPSYSCONF_SERVERDIR, interface);

    if(file_exists(stringbuffer_getstring(sb))) {

        fname = xstrdup(stringbuffer_getstring(sb));
        stringbuffer_destroy(sb);
        return fname;

    }

    /* otherwise there is no configuration file and this is
     * bad. signal a fatal error. */

    FATAL_MESSAGE("configuration file missing -- looked for default.conf and <interface>.conf. exiting...");
    exit(1); /* get rid of compiler warning. */
}

/* set variable. */
static int server_conf_set_variable(server_conf_t *server_conf, arg_symbol_t var_symbol, const char *var_value)
{
    switch(var_symbol) {

    case SERVER_VAR_RENEW_PERCENT:

        if(!is_unsigned_numeric(var_value))  {
            ERROR_MESSAGE("value not unsigned numeric as expected : %s\n", var_value);
            return 1;
        }

        sscanf(var_value, "%"SCNu16, &server_conf->default_renew_percent);

        if(server_conf->default_renew_percent > 100) { /* check if the percent is higher than it should be. */
            FATAL_MESSAGE("value specified for default-renew-percent is an illegal percent (over 100)");
        }

        break;

    case SERVER_VAR_REBIND_PERCENT:

        if(!is_unsigned_numeric(var_value))  {
            ERROR_MESSAGE("value not unsigned numeric as expected : %s", var_value);
            return 1;
        }

        sscanf(var_value, "%"SCNu16, &server_conf->default_rebind_percent);

        if(server_conf->default_renew_percent > 100) { /* check if the percent is higher than it should be. */
            FATAL_MESSAGE("value specified for default-rebind-percent is an illegal percent (over 100)");
        }

        break;

    case SERVER_VAR_POLL_TIMEOUT:

        if(!is_unsigned_numeric(var_value))  {
            ERROR_MESSAGE("value not unsigned numeric as expected : %s\n", var_value);
            return 1;
        }

        sscanf(var_value, "%"SCNu16, &server_conf->poll_timeout);
        break;

    default:
        
        ERROR_MESSAGE("unable to understand assignment value of : %s", var_value);
        return 1;
    }

    return 0;
}

/* load server options. */
static int server_conf_load_options(server_conf_t *sc)
{
    conf_t *conf;
    list_t *directives;
    directive_t *directive;

    conf = conf_create(commands, sc->conf_file);

    if(conf_compile_directives(conf) == CONF_ERROR) {
        ERROR_MESSAGE("error reading conf file: %s at line %d", sc->conf_file, conf_get_line_no(conf));
        conf_destroy(conf);
        return 1;
    }

    directives = conf_get_directives(conf);

    list_rewind(directives);
    while((directive = list_next(directives)) != NULL) {

        /* read in the compiled directives. */
        if(directive_handlers[directive->command_code](sc, directive, GROUP_NULL, NULL))
            return 1;
    }

    return 0;
}

/* * * * * * * * * * * * 
 * directive handlers. *
 * * * * * * * * * * * */

/* variable set handler. */
static int directive_set_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data)
{
    list_t *args;
    arg_symbol_t *var_symbol;
    char *var_value;

    args = directive->arguments;

    var_symbol = list_get_by_index(args, 0);
    var_value = list_get_by_index(args, 2);

    if(server_conf_set_variable(server_conf, *var_symbol, var_value)) {
        ERROR_MESSAGE("configuration: error unable to set variable %s to %s", var_strings[*var_symbol], var_value);
        return 1;
    }

    return 0;
}

static int directive_lease_handler_proc(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data,
                                        list_t *option_list)
{
    directive_t *lower_directive;
    list_t *lease_directives, *args;

    /* arguments. */
    args = directive->arguments;

    /* parse lower directives. */
    lease_directives = list_get_by_index(args, 2);

    list_rewind(lease_directives);
    while((lower_directive = list_next(lease_directives)) != NULL) {

        switch(lower_directive->command_code) {

        case DIRECTIVE_OPTION:

            if(directive_option_handler(server_conf, lower_directive, group_type, option_list))
                return 1;

            break;

        default:
            ERROR_MESSAGE("illegal directive specified in a range lease definition");
        }
    }

    return 0;
}

/* lease range handler */
static int directive_range_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data)
{
    lease_definition_t *lease_def;
    ip_addr_t bottom_address, top_address;
    list_t *args, *address_pair, *option_list;
    uint32_t lease_expiry = 0;
    uint32_t renew_time = 0;
    uint32_t rebind_time = 0;

    /* arguments. */
    args = directive->arguments;

    /* get address pair, and create pair. */
    bottom_address = *(ip_addr_t *)list_first(args);
    top_address    = *(ip_addr_t *)list_second(args);

    /* make sure address range is correct. */
    if(ntohl(bottom_address) >= ntohl(top_address)) {
        ERROR_MESSAGE("range lease defined with bottom address higher or equal to top address.");
        return 1;
    }

    address_pair = list_create();
    list_add(address_pair, &top_address);
    list_add_to_end(address_pair, &bottom_address);

    /* create option handler by calling the general lease handler. */
    option_list = list_create();
    directive_lease_handler_proc(server_conf, directive, GROUP_LEASE_DEF, group_data, option_list);

    lease_def = lease_definition_create(NULL, LEASE_RANGE_ADDRESS, address_pair, option_list, lease_expiry, renew_time, rebind_time);
    list_add(server_conf->lease_defs, lease_def);

    /* clean up. */
    list_destroy(address_pair, NULL);

    return 0;
}

/* hostname handler. */
static int directive_hostname_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data)
{
    lease_constraint_t *constraint;
    list_t *args, *option_list;
    char *hostname;
    ip_addr_t *ip_address;
    lease_definition_t *lease_def;
    uint32_t lease_expiry = 0;
    uint32_t renew_time = 0;
    uint32_t rebind_time = 0;

    /* arguments. */
    args = directive->arguments;

    /* first argument is the hostname. */
    hostname = list_first(args);

    if((constraint = lease_constraint_create(LEASE_CONSTRAINT_HOSTNAME, hostname)) == NULL) {
        ERROR_MESSAGE("illegal hostname constraint specified: %s", hostname);
        return 1;
    }

    ip_address = list_second(args);
    option_list = list_create();
    directive_lease_handler_proc(server_conf, directive, GROUP_LEASE_DEF, group_data, option_list);

    lease_def = lease_definition_create(constraint, LEASE_SINGLE_ADDRESS, ip_address, option_list, 
                                        lease_expiry, renew_time, rebind_time);
    list_add(server_conf->lease_defs, lease_def);

    return 0;
}

/* hardware address handler. */
static int directive_hw_address_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data)
{
    lease_constraint_t *constraint;
    list_t *args, *option_list;
    eth_addr_t *eth_addr;
    char *eth_addr_string;
    ip_addr_t *ip_address;
    lease_definition_t *lease_def;
    uint32_t lease_expiry = 0;
    uint32_t renew_time = 0;
    uint32_t rebind_time = 0;

    /* arguments. */
    args = directive->arguments;

    /* first argument is the hostname. */
    eth_addr = list_first(args);

    if((constraint = lease_constraint_create(LEASE_CONSTRAINT_HARDWARE_ADDRESS, eth_addr)) == NULL) {
        eth_addr_string = eth_addr_to_string(*eth_addr);
        ERROR_MESSAGE("illegal mac address constraint specified: %s", eth_addr_string);
        xfree(eth_addr_string);
        return 1;
    }

    ip_address = list_second(args);
    option_list = list_create();
    directive_lease_handler_proc(server_conf, directive, GROUP_LEASE_DEF, group_data, option_list);

    lease_def = lease_definition_create(constraint, LEASE_SINGLE_ADDRESS, ip_address, option_list, 
                                        lease_expiry, renew_time, rebind_time);
    list_add(server_conf->lease_defs, lease_def);

    return 0;
}

/* lease option handler. */
static int directive_option_handler(server_conf_t *server_conf, directive_t *directive, int group_type, void *group_data)
{
    dhcp_opt_t *option = NULL; /* get rid of compiler warning. */
    list_t *option_list, *option_data_list;
    char *option_name;
    uint8_t i;
    const char **option_strings;

    if(group_type == GROUP_NULL) {
        ERROR_MESSAGE("dhcp option setting found outside of lease definition");
        return 1;
    }

    option_name = list_first(directive->arguments);
    option_data_list = list_get_by_index(directive->arguments, 2);
    option_strings = dhcp_option_conf_string_get_array();

    for(i = 0; i < MAX_OPTIONS_HANDLED; i++) {

        if(!strcmp(option_strings[i], option_name)) {

            if((option = dhcp_opt_create_from_user_string(i, option_data_list)) == NULL) {
                ERROR_MESSAGE("illegal specification for option %s", option_name);
                return 1;
            }

            break;
        }
    }

    if(i == MAX_OPTIONS_HANDLED) {
        ERROR_MESSAGE("unrecognized dhcp option specified by user config: %s", option_name);
        return 1;
    }


    option_list = group_data;
    list_add_to_end(option_list, option);

    return 0;
}

/* * * * * * * * * * *
 * public interface  *
 * * * * * * * * * * */

server_conf_t *server_conf_create(const char *interface)
{
    server_conf_t *sc;

    sc = xcalloc(sizeof(server_conf_t));
    sc->interface = interface;
    sc->conf_file = get_conf_options_fname(interface);
    
    /* setup lease definitions. */
    sc->lease_defs = list_create();

    /* set defaults. */
    sc->default_rebind_percent = SERVER_DEFAULT_REBIND_PERCENT;
    sc->default_renew_percent  = SERVER_DEFAULT_RENEW_PERCENT;
    sc->poll_timeout           = SERVER_DEFAULT_POLL_TIMEOUT;

    if(server_conf_load_options(sc)) {
        server_conf_destroy(sc);
        return NULL;
    }

    return sc;
}

void server_conf_destroy(server_conf_t *sc)
{
    /* not implemented yet. */
    return;
}
